---------------------------------------------------------------------------
-- This VHDL file was developed by Daniel Llamocca (2013).  It may be
-- freely copied and/or distributed at no cost.  Any persons using this
-- file for any purpose do so at their own risk, and are responsible for
-- the results of such use.  Daniel Llamocca does not guarantee that
-- this file is complete, correct, or fit for any particular purpose.
-- NO WARRANTY OF ANY KIND IS EXPRESSED OR IMPLIED.  This notice must
-- accompany any copy of this file.
--------------------------------------------------------------------------

-- This hardware only multiplies two numbers of 4 bits
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

entity my_wallace_4bit is
	port (A, B: in std_logic_vector (3 downto 0);
	      P : out std_logic_vector (7 downto 0));
end my_wallace_4bit;

architecture structure of my_wallace_4bit is

	component fulladd
		port( cin, x, y : in std_logic;
				s, cout   : out std_logic);
	end component;
	
	type my_array is array (natural range <>, natural range <>) of std_logic;
	signal m: my_array(3 downto 0, 3 downto 0);
	signal c: my_array(2 downto 0, 4 downto 0);
	signal s: my_array(2 downto 0, 4 downto 0);

begin
	
	gi: for i in 0 to 3 generate -- along rows
	      gj: for j in 0 to 3 generate -- along columns
					   m(i,j) <= A(i) and B(j);			         
				 end generate;
		 end generate;
	
	c(0,0) <= '0'; c(1,0) <= '0'; c(2,0) <= '0';
	s(0,4) <= c(0,4); s(1,4) <= c(1,4); s(2,4) <= c(2,4);
	
-- First Layer	
	f00: fulladd port map (cin => c(0,0), x => m(0,1), y => m(1,0), s => s(0,0), cout => c(0,1));
	f01: fulladd port map (cin => c(0,1), x => m(1,1), y => m(2,0), s => s(0,1), cout => c(0,2));
	f02: fulladd port map (cin => c(0,2), x => m(2,1), y => m(3,0), s => s(0,2), cout => c(0,3));
	f03: fulladd port map (cin => c(0,3), x => m(3,1), y => '0', s => s(0,3), cout => c(0,4));
	
-- Second Layer
	f10: fulladd port map (cin => c(1,0), x => m(0,2), y => s(0,1), s => s(1,0), cout => c(1,1));
	f11: fulladd port map (cin => c(1,1), x => m(1,2), y => s(0,2), s => s(1,1), cout => c(1,2));
	f12: fulladd port map (cin => c(1,2), x => m(2,2), y => s(0,3), s => s(1,2), cout => c(1,3));
	f13: fulladd port map (cin => c(1,3), x => m(3,2), y => s(0,4), s => s(1,3), cout => c(1,4));
	
-- Third Layer
	f20: fulladd port map (cin => c(2,0), x => m(0,3), y => s(1,1), s => s(2,0), cout => c(2,1));
	f21: fulladd port map (cin => c(2,1), x => m(1,3), y => s(1,2), s => s(2,1), cout => c(2,2));
	f22: fulladd port map (cin => c(2,2), x => m(2,3), y => s(1,3), s => s(2,2), cout => c(2,3));
	f23: fulladd port map (cin => c(2,3), x => m(3,3), y => s(1,4), s => s(2,3), cout => c(2,4));
	
	P(0) <= m(0,0);
	P(1) <= s(0,0);
	P(2) <= s(1,0);
	P(3) <= s(2,0);
	P(4) <= s(2,1);
	P(5) <= s(2,2);
	P(6) <= s(2,3);
	P(7) <= s(2,4);
	
end structure;

